/*
 * syslinux.c
 *
 * Creates a Linux bootable MS-DOS floppy
 *
 * Copyright (C) 1994 H. Peter Anvin
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>

extern char boot_sector[];
extern char ldlinux_sys[];
extern unsigned int ldlinux_len;

#define BOOTSECT_LEN    512         /* Length of the boot sector */
#define RETRIES         4           /* Make 4 read/write attempts */

#define DEFCMD_PTR      0x20        /* Compilation constants for */
#define DEFCMD_MAXLEN   0x22        /* LDLINUX.SYS */
#define FAT_MAXSIZE     0x24

typedef unsigned int word;
typedef unsigned char byte;
typedef unsigned long dword;

#define sysword(OFFS) ( *( (word *) (ldlinux_sys+(OFFS)) ) )

char *default_cmd = "vmlinux";      /* Default kernel image */
char *auto_str = " auto";           /* Add to default command, always */

struct bootsector
{
    byte jump[3];
    byte oemname[8];
    word bytespersec;
    byte secperclust;
    word ressectors;
    byte fats;
    word rootdirents;
    word sectors;
    byte media;
    word fatsecs;
    word secpertrack;
    word heads;
    dword hiddensecs;
    dword hugesectors;
    byte drivenumber;
    byte reserved1;
    byte bootsignature;
    dword volumeid;
    byte volumelabel[11];
    byte filesystype[8];
    byte bootstrap_code[448];
    word boot_magic;
};

struct diskio
{
    unsigned long sector;
    unsigned count;
    void far *buffer;
};

/*
 * read_boot_sect:  read the boot sector of the specified drive into the
 *                  given buffer, using the DOS function INT 25h
 */
int read_boot_sect(int drive, void *buffer, int huge_disk)
{
    int retry_ctr, err;
    struct diskio huge_buf;
    unsigned sectors;

    if ( huge_disk )
    {
        huge_buf.buffer = buffer;
        huge_buf.sector = 0;
        huge_buf.count = 1;
        buffer = &huge_buf;
        sectors = 0xFFFF;
    }
    else
        sectors = 1;            /* Read one sector */

    err = 0;
    for ( retry_ctr = RETRIES ; retry_ctr ; retry_ctr-- )
    {
        _BX = (unsigned) buffer;
        _CX = sectors;
        _DX = 0;                    /* Sector zero = boot sector */
        _AL = drive;
        geninterrupt(0x25);
        asm pop ax;                 /* Flags left on the stack */
        asm sti;                    /* Make sure interrupts are on */
        asm jnc read_ok;
    }
    err = 1;                        /* Operation failed */
read_ok:
    return err;
}

/*
 * write_boot_sect:  read the boot sector of the specified drive into the
 *                   given buffer, using the DOS function INT 26h
 */
int write_boot_sect(int drive, void *buffer, int huge_disk)
{
    int retry_ctr, err, flg;
    struct diskio huge_buf;
    unsigned sectors;

    if ( huge_disk )
    {
        huge_buf.buffer = buffer;
        huge_buf.sector = 0;
        huge_buf.count = 1;
        buffer = &huge_buf;
        sectors = 0xFFFF;
    }
    else
        sectors = 1;            /* Write one sector */


    err = 0;
    for ( retry_ctr = RETRIES ; retry_ctr ; retry_ctr-- )
    {
        _BX = (unsigned) buffer;
        _CX = sectors;
        _DX = 0;                    /* Sector zero = boot sector */
        _AL = drive;
        geninterrupt(0x26);
        asm pop ax;                 /* Flags left on the stack */
        asm sti;                    /* Make sure interrupts are on */
        asm jnc write_ok;
    }
    err = 1;                        /* Operation failed */
write_ok:
    return err;
}

/*
 * disk_is_huge:    Check if a disk is larger than 64K sectors (32 Mb)
 */
int disk_is_huge(int drive, int *sectsize)
{
    unsigned secperclust, bytespersec, clusters;
    unsigned long sectors;

    _DL = drive+1;
    _AH = 0x36;
    geninterrupt(0x21);
    *sectsize = _CX;
    asm mul dx;

    return _DX;
}

/*
 * usage:   Complain to the user, and exit
 */
void usage(void)
{
    fprintf(stderr, "Usage:\n"
                    "         SYSLINUX drive: [default command line]\n"
                    "\n"
                    "The default command line should begin with the name\n"
                    "of the default kernel image.  If no command line is\n"
                    "specified, the default kernel name is set to "
                    "\"%s\".\n", default_cmd);
    exit(1);
}

/*
 * Main program
 *
 */
int main(int argc, char *argv[])
{
    int drive, huge_disk, sectsize;
    char *cmd_ptr;
    int cmd_maxlen, cmd_len;
    struct bootsector *orig_sect;
    static char *ld_name = "?:\LDLINUX.SYS";
    FILE *ld;
    int i;


    /* If no drive specified */
    if ( argc < 2 || strlen(argv[1]) != 2 || argv[1][1] != ':' )
        usage();

    drive = argv[1][0] | 0x20;
    if ( drive < 'a' || drive > 'z' )
        usage();

    drive -= 'a';
    huge_disk = disk_is_huge(drive, &sectsize);

    orig_sect = malloc(sectsize);
    if ( !orig_sect )
    {
        fprintf(stderr, "Error: could not malloc %d bytes\n", sectsize);
        exit(1);
    }

    /* Read in the original boot sector */
    if ( read_boot_sect(drive, orig_sect, huge_disk) )
    {
        fprintf(stderr, "Error: could not read original boot sector\n");
        exit(1);
    }

    /* Make sure we're dealing with something sane here */
    if ( orig_sect->bytespersec != 512 )
    {
        fprintf(stderr, "Error: Bytes/sector is not 512\n");
        exit(1);
    }
    if ( orig_sect->boot_magic != 0xAA55 )
    {
        fprintf(stderr, "Error: Boot sector signature not found on disk\n");
        exit(1);
    }
    if ( orig_sect->fats * orig_sect->fatsecs * orig_sect->bytespersec >
         sysword(FAT_MAXSIZE) )
    {
        fprintf(stderr, "Error: FAT too large (more than %d bytes)\n",
                sysword(FAT_MAXSIZE));
        exit(1);
    }
    if ( orig_sect->bootsignature == 0x29 &&
         memcmp(orig_sect->filesystype, "FAT12   ", 8) )
    {
        fprintf(stderr, "Error: Filesystem is not type FAT12\n");
        exit(1);
    }

    /* Very well, copy the thing */
    memcpy( &boot_sector[0x0B], &(orig_sect->bytespersec), 0x33 );

    /* Get a pointer to default command line */
    cmd_ptr = ldlinux_sys + sysword(DEFCMD_PTR);
    cmd_maxlen = sysword(DEFCMD_MAXLEN) - strlen(auto_str);

    /* Compute the length of the specified command line */
    if ( argc < 3 )
    {
        cmd_len = strlen(default_cmd);
    }
    else
    {
        cmd_len = argc-3;
        for ( i = 2 ; i < argc ; i++ )
            cmd_len += strlen(argv[i]);
    }

    if ( cmd_len > cmd_maxlen )
    {
        fprintf(stderr,
                "Error: default command line too long (max %d bytes)\n",
                cmd_maxlen);
        exit(1);
    }

    /* Copy the default command line */
    if ( argc < 3 )
    {
        fprintf(stderr, "Setting default kernel name to \"%s\"\n", default_cmd);
        strcpy(cmd_ptr, default_cmd);
        cmd_ptr = strchr(cmd_ptr, '\0');
    }
    else
    {
        for ( i = 2 ; i < argc ; i++ )
        {
            strcpy(cmd_ptr, argv[i]);
            cmd_ptr = strchr(cmd_ptr, '\0');
            *(cmd_ptr++) = ' ';
        }
        *(--cmd_ptr) = '\0';
    }
    strcpy(cmd_ptr, auto_str);

    /* Write out the new boot sector */
    if ( write_boot_sect(drive, boot_sector, huge_disk) )
    {
        fprintf(stderr, "Error: could not write new boot sector\n");
        exit(1);
    }

    /* Write LDLINUX.SYS */
    *ld_name = drive+'A';
    ld = fopen(ld_name, "wb");
    if ( !ld )
    {
        perror("Writing LDLINUX.SYS");
        exit(1);
    }
    if ( fwrite(ldlinux_sys,1,ldlinux_len,ld) < ldlinux_len )
    {
        perror("Writing LDLINUX.SYS");
        fclose(ld);
        exit(1);
    }
    fclose(ld);

    fprintf(stderr, "The Linux boot system was successfully installed.\n");
    return 0;
}

